/* --------------------------------------------------------------- */
/* -- Copyright (c) 1994, 1995 Computer Sciences Department,    -- */
/* -- University of Wisconsin-Madison, subject to the terms     -- */
/* -- and conditions given in the file COPYRIGHT.  All Rights   -- */
/* -- Reserved.                                                 -- */
/* --------------------------------------------------------------- */

/*
 * $Id: pin.c,v 1.110 1996/09/04 18:59:36 nhall Exp $
 */
#define SM_SOURCE
#define FACT_PIN_C

#ifdef __GNUG__
#pragma implementation "fact_pin.h"
#endif

#include <sm_int.h>
#include <sm.h>
#include <fact_pin.h>
#include <prologue.h>
#include "smstats.h"

#ifdef CHEAP_RC

/* delegate not used for DO_GOTO; have to delete the _err */
#define RETURN_RC( _err) \
    { 	w_rc_t rc = rc_t(_err); if(_err) delete _err; return  rc; }

#else

/* delegate is used in debug case */
#define RETURN_RC( _err) \
    { 	w_rc_t rc = rc_t(_err); return  rc; }

#endif

// TO DO... Should be removed later
inline
latch_mode_t lock_to_latch(lock_mode_t m)
{
    switch(m) {
    case SH:
    case UD:
    case NL:
        return LATCH_SH;
    case EX:
        return LATCH_EX;

    default:
        W_FATAL(eNOTIMPLEMENTED);
    }
    return LATCH_NL; // never gets here
}


fact_pin_i::~fact_pin_i()
{ 
    unpin();

    // unpin() actuall calls unfix on these pages and unfix does
    // everything the destructor does, but to be safe we still call
    // the destructor
    _data_page().destructor();
}

inline void
fact_pin_i::_set_lsn()
{
#ifdef DEBUG
    xct()->flush_logbuf();
    _data_lsn = _data_page().lsn();
#endif /*DEBUG*/
}

rc_t fact_pin_i::pin(const rid_t rid, smsize_t start, lock_mode_t lmode)
{
    SM_PROLOGUE_RC(fact_pin_i::pin, in_xct, 2);
    if (lmode != SH && lmode != UD && lmode != EX && lmode != NL)
	return RC(eBADLOCKMODE);
    W_DO(_pin(rid, start, lmode, serial_t::null));
    _set_lsn();
    return RCOK;
}

void fact_pin_i::unpin()
{
    if (pinned()) {
	_data_page().unfix();  
	_flags = pin_empty;
	_rec = NULL;

	incr_static_unpin_count();
    }
}

void fact_pin_i::set_ref_bit(int value)
{
    if (pinned()) {
        _data_page().set_ref_bit(value);  
    }
}

rc_t fact_pin_i::repin(lock_mode_t lmode)
{
    SM_PROLOGUE_RC(fact_pin_i::repin, in_xct, 2);
    if (lmode != SH && lmode != UD && lmode != EX) return RC(eBADLOCKMODE);
    DBG(<<" repin " << this->_rid);
    W_DO(_repin(lmode));
    return RCOK;
}

const char* fact_pin_i::data_page_data()
{
    _check_lsn();
    if (!pinned()) return NULL;
    return (const char*) &(_data_page().persistent_part());
}

const lsn_t& fact_pin_i::_get_data_lsn() const { return _data_page().lsn(); }

void fact_pin_i::_init_constructor()
{
    w_assert3(sizeof(_data_page_alias) == sizeof(fixed_p));

    _flags = pin_empty;
    _rec = NULL;
    _lmode = NL; 
    new (_data_page_alias) fixed_p();
}

rc_t fact_pin_i::_pin(const rid_t rid, smsize_t start, lock_mode_t lmode, const serial_t& verify)
{
    w_error_t*   err=0;	// must use rather than rc_t due to HP_CC_BUG_2
    bool pin_page = FALSE;	// TRUE indicates page needs pinning

    w_assert3(lmode == SH || lmode == EX || lmode == UD ||
	    lmode == NL /*for scan.c*/ );

    DBG(<<"enter _pin");
    if (pinned()) {
	DBG(<<"pinned");
   
	/*
	 * If the page for the new record is not the same as the
	 * old (or if we need to get a lock),
	 * then unpin the old and get the new page.
	 * If we need to get the lock, then we must unfix since
	 * we may block on the lock.  We may want to do something
	 * like repin where we try to lock with 0-timeout and
	 * only if we fail do we unlatch.
	 */
	if (rid.pid != _rid.pid || lmode != NL) {
	    _data_page().unfix();  
	    pin_page = TRUE;
	    incr_static_unpin_count();
	}
    } else {
	DBG(<<"not pinned");
	pin_page = TRUE;
    }

    // aquire lock only lock is requested
    if (lmode != NL) {
	DBG(<<"acquiring lock");
	W_DO_GOTO(err, lm->lock(rid, lmode, t_long, WAIT_SPECIFIED_BY_XCT));
	DBG(<<"lock is acquired");
	_lmode = lmode;
    } else {
	// we trust the caller and therefore can do this
	if (_lmode == NL) _lmode = SH;
    }
    w_assert3(_lmode > NL); 

    if (pin_page) {
	W_DO_GOTO(err, fl->locate_page(rid, _data_page(), 
                                            lock_to_latch(_lmode)));
	incr_static_pin_count();
    }

    _data_page().retrieve_scan_rec(rid.slot, _rec);
    if (_rec == NULL) goto failure;
    if (start > 0 && start >= body_size()) {
	err = RC(eBADSTART).delegate();
	goto failure;
    }

    _flags = pin_rec_pinned;
    _rid = rid;

    /*
     * See if the record is small or large.  Record number zero
     * is a special header record and is considered small
     */
     _start = 0;
     _len = body_size()-_start;

/* success: */
    _set_lsn();
    return RCOK;
  
failure:
    if (pin_page) {
	_data_page().unfix();  
	incr_static_unpin_count();
    }
    _flags = pin_empty;
    RETURN_RC(err);
}

#ifndef MULTI_SERVER
#define old_value /*not used*/
#endif
rc_t fact_pin_i::_repin(lock_mode_t lmode,
			int* old_value /* temporary: for degugging only */)
#undef old_value
{
    w_error_t*   err=0;	// must use rather than rc_t due to HP_CC_BUG_2

    w_assert3(lmode == SH || lmode == UD || lmode == EX);
    // acquire lock if current one is not strong enough
    // TODO: this should probably use the lock supremam table
    if (_lmode < lmode) {
	DBG(<<"acquiring lock");
	// see if we can get the lock without blocking
#ifndef MULTI_SERVER
	rc_t rc = lm->lock(_rid, lmode, t_long, 0);
#else
	rc_t rc = lm->lock(_rid, lmode, t_long, 0, TRUE);
#endif
	if (rc) {
	    if (rc.err_num() == eMAYBLOCK || rc.err_num() == eLOCKTIMEOUT) {
	        // we would block, so unpin is necessary, then get lock
	        if (pinned()) unpin();
#ifndef MULTI_SERVER
	        W_DO_GOTO(err, lm->lock(_rid, lmode, t_long,
				WAIT_SPECIFIED_BY_XCT));
#else
	        W_DO_GOTO(err, lm->lock(_rid, lmode, t_long,
				WAIT_SPECIFIED_BY_XCT, FALSE, old_value));
#endif
	    } else {
		W_DO_GOTO(err, rc.reset());
	    }
	}
	DBG(<<"lock is acquired");
	_lmode = lmode;
    }	

    if (pinned()) {
	w_assert3(_data_page().is_fixed());  

	// upgrade to an EX latch if all we had before was an SH latch
	if (_data_page().latch_mode() != lock_to_latch(_lmode)) {
	    w_assert3(_data_page().latch_mode() == LATCH_SH);
	    w_assert3(_lmode == EX || _lmode == UD);

	    bool would_block = false;  // was page unpinned during upgrade
	    W_DO_GOTO(err, _data_page().upgrade_latch_if_not_block(would_block));
	    if (would_block) {
		unpin();
	    } else {
		w_assert3(_data_page().latch_mode() == lock_to_latch(_lmode));
	    }
	}
    }

    if (pinned()) {
#ifdef DEBUG
	// make sure record is where it's supposed to be
	void* tmp;
	W_DO_GOTO(err, _data_page().retrieve_rec(_rid.slot, tmp) );
	w_assert3(tmp == _rec);
#endif
    } else { 
	// find the page we need and latch it
	W_DO_GOTO(err, fl->locate_page(_rid, _data_page(), lock_to_latch(_lmode)));
	W_DO_GOTO(err, _data_page().retrieve_rec(_rid.slot, _rec));
	w_assert3(_rec);
	if (_start > 0 && _start >= body_size()) {
	    err = RC(eBADSTART).delegate();
	    goto failure;
	}
	_flags = pin_rec_pinned;

	/*
	 * See if the record is small or large.  Record number zero
	 * is a special header record and is considered small
	 */
        _start = 0;
	_len = body_size()-_start;
	incr_static_pin_count();
    }

/* success: */
    _set_lsn();
    return RCOK;
  
failure:
    _flags = pin_empty;
    RETURN_RC(err);
}

const char* fact_pin_i::body()
{
    if ( !pinned() )
	return NULL;

    // must be a small record
    _check_lsn();
    w_assert3(is_aligned(_rec));
    return (char *)_rec;
}

#ifdef DEBUG
void fact_pin_i::_set_lsn_for_scan()
{
    _data_lsn = _data_page().lsn();
}
#endif /*DEBUG*/
