/* --------------------------------------------------------------- */
/* -- Copyright (c) 1994, 1995 Computer Sciences Department,    -- */
/* -- University of Wisconsin-Madison, subject to the terms     -- */
/* -- and conditions given in the file COPYRIGHT.  All Rights   -- */
/* -- Reserved.                                                 -- */
/* --------------------------------------------------------------- */

/*
 * $Id: pin.h,v 1.70 1996/09/05 22:59:52 nhall Exp $
 */
#ifndef FACT_PIN_H
#define FACT_PIN_H

#ifdef __GNUG__
#pragma interface
#endif

/***********************************************************************
   The fact_pin_i class (located in pin.h) is used to pin ranges of bytes in
   a record.  The amount pinned (of the record body) can be determined
   with the start_byte() and length() functions.  Access to the pinned
   region is via the body() function.  The header is always pinned if
   any region is pinned and can be accessed via hdr().
   
   ~fact_pin_i() will unpin the record.  Pin() and unpin() can also 
   be used to change which record is pinned (pin() will unpin the
   currently pinned record()).
  
   For large records, data pages will not actually be pinned until
   body() is called.  Therefore, to just read record
   headers, pinning with start 0 will not cause any additional IO.
  
   The repin function efficiently re-pins a previously unpinned record
   and efficiently repins a record even while it is pinned.  This is
   useful after append_rec and truncate_rec calls to repin the record
   since its location may have changed.

   NOTE ON LOCK MODE PARAMETERS:
      The fact_pin_i, pin, repin functions all take a lock mode parameter
      that specifies how the record should initially be locked.  The
      options are SH and EX.  EX should be used when
      the pinned record will be eventually updated (throuh update_rec,
      unpdate_rec_hdr, append_rec, or truncate_rec).  Using EX in these
      cases will improve performance and reduce the risk of deadlock,
      but is not necessary for correctness.

   WARNING:
     The fact_pin_i structure for a pinned record is no longer valid after
     any append, truncate, create operation for ANY record on the page
     that is pinned.  To enforce this a debugging check is made that
     compares the page's current lsn with its value when the record was
     pinned.  Therefore, update_rec calls must also have a repin call
     performed.

   For efficiency (no lrid lookups) and to avoid repinning, the
   ss_m::update_rec and ss_m::update_rec_hdr functions are also
   provided by fact_pin_i.  These can be called on any pinned record
   regardless of where and how much is pinned.  If a fact_pin_i was
   previously pinned and then upinned, a call to
   fact_pin_i::update_rec[_hdr] will temporarily repin the record and then
   unpin it.  Therefore, after any fact_pin_i call that updates the record,
   the state of the fact_pin_i (either pinned or not) remains the same.

 **********************************************************************/

struct fixed_p;

class fact_pin_i : public smlevel_top {
    friend class scan_fact_i;
    friend class scan_bitmap_i;
    friend class scan_plain_i;
public:
    enum flags_t { 
	pin_empty		= 0x0,
	pin_rec_pinned		= 0x01
    };
    
    NORET		fact_pin_i() {_init_constructor();}

    NORET		~fact_pin_i();

    // these functions pin portions of a record beginning at start
    // the actuall location pinned (returned by start_byte), may
    // be <= start.
    rc_t		pin(
	const rid_t	    rid,
	smsize_t	    start,
	lock_mode_t 	    lmode = SH);

    void   		unpin();

    // set_ref_bit sets the reference bit value to use for the buffer
    // frame containing the currently pinned body page when the page
    // is unpinned.  A value of 0 is a "hate" hint indicating that
    // the frame can be reused as soon as necessary.  By default,
    // a value of 1 is used indicating the page will be cached 
    // until at least 1 sweep of the buffer clock hand has passed.
    // Higher values cause the page to remain cached longer.
    void		set_ref_bit(int value);

    // repin is used to efficiently repin a record after its size
    // has been changed, or after it has been unpinned.
    rc_t    		repin(lock_mode_t lmode = SH);

    // is something currently pinned
    bool  		pinned() const     
		{ return _flags & pin_rec_pinned; }
    // is the entire record pinned
    bool  		pinned_all() const 
		{ return pinned() && _start==0 && _len==body_size();}

    // return true if pinned *and* pin is up-to-date with the LSN on
    // the page.  in other words, verify that the page has not been
    // updated since it was pinned by this fact_pin_i
    bool		up_to_date() const
		{ return pinned() && (_data_lsn == _get_data_lsn());}

    smsize_t   start_byte() const { _check_lsn(); return _start;}
    smsize_t   length() const     { _check_lsn(); return _len;}
    smsize_t   body_size() const  { _check_lsn(); return _record_size;}

    const rid_t&     rid() const { _check_lsn(); return _rid;}
    const char*      body();

    const char*      data_page_data();

private:

    void        _set_record_size( smsize_t size ) { _record_size = size ; } ;  
    void        _init_constructor(); // companion to constructor
    rc_t         _pin_data();
    rc_t        _pin(const rid_t rid, smsize_t start, lock_mode_t m, const serial_t& verify);
    rc_t	_repin(lock_mode_t lmode, int* old_value = 0);

    fixed_p*    _get_data_page_no_lsn_check() const {return pinned() ?
			               (fixed_p*)_data_page_alias : 0;}
    fixed_p*    _get_data_page() const { _check_lsn(); return _get_data_page_no_lsn_check();}

    // NOTE: if the _check_lsn assert fails, it usually indicates that
    // you tried to access a pinned record after having updated the
    // record, but before calling repin.
    // The _set_lsn() function is used to reset the lsn to the page's
    // new value, after an update operation.
    // void _check_lsn() const {w_assert3(up_to_date());}
    void _check_lsn() const {}
    void _set_lsn();
    void _set_lsn_for_scan() // used in scan.c
#ifndef DEBUG
	{}  	// nothing to do if not debugging
#endif
	; 	// defined in scan.c

    const lsn_t& _get_data_lsn() const;

    rid_t	_rid;
    smsize_t	_len;
    smsize_t    _record_size ; 
    smsize_t	_start;
    void*       _rec ; 
    uint4 	_flags;  // this cannot be flags_t since it uses
    // | to generate new flags not in the enum 
    // _hdr_lsn is used to record the lsn on the page when
    // the page is pinned.  When compiled with -DDEBUG, all fact_pin_i
    // operations check that the hdr page's _lsn1 has not changed
    // (ie. to verify that the pinned record has not moved)
    lsn_t	_data_lsn;
    lock_mode_t _lmode;  // current locked state

    /*
     *	Originally fact_pin_i contained the _hdr_page and _hdr_page data
     *	members commented out below.  This required that users #include
     *	sm_int.h (ie. the whole world), generating large .o's.
     *	Instead, we have the corresponding "alias" byte arrays and
     *	member functions which cast these to the correct page type.
     *	Only pin.c uses these functions.  This greatly reduces the
     *	number of .h files users need to include.
     *
     *  Asserts in fact_pin_i constructors verify that the _alias members
     *  are large enough to hold file_p and lgdata_p. 
     */
    fixed_p&	_data_page() const;
    char        _data_page_alias[20];

    /*
     * Statistics for record pinning
     */
    static void incr_static_pin_count()	  {
		smlevel_0::stats.rec_pin_cnt++;
	}
    static void incr_static_unpin_count() {
		smlevel_0::stats.rec_unpin_cnt++;
	}

    // disable
    fact_pin_i(const fact_pin_i&);
    fact_pin_i& operator=(const fact_pin_i&);

};

inline fixed_p&	fact_pin_i::_data_page() const
			{return *(fixed_p*)_data_page_alias;}

#endif	// FACT_PIN_H 
