/* --------------------------------------------------------------- */
/* -- Copyright (c) 1994, 1995 Computer Sciences Department,    -- */
/* -- University of Wisconsin-Madison, subject to the terms     -- */
/* -- and conditions given in the file COPYRIGHT.  All Rights   -- */
/* -- Reserved.                                                 -- */
/* --------------------------------------------------------------- */

/*
 *  $Id: fixed.h,v 1.85 1996/09/04 18:59:34 nhall Exp $
 */
#ifndef FIXED_H
#define FIXED_H

#ifdef __GNUG__
#pragma interface
#endif

#include "fixed_s.h"
#include "fixed_dir.h"
#include "bitmap.h"

/*--------------------------------------------------------------*
 *  class fixed_p                                               *
 *                                                              * 
 * Fixed length record page class. It doesn't contain any slot, * 
 * header and other record related space overhead associated    *
 * with slotted pages.                                          *
 *                                                              *
 *      Initial Version 11/08/96 -- Karthikeyan Ramasamy        *
 *                                                              *
 *--------------------------------------------------------------*/

class fixed_p : public page_p {
  
   public:

     // free space on fixed_p is page_p 
     enum { data_sz = page_p::data_sz - sizeof(fixed_p_hdr_t) };
  
     MAKEPAGE(fixed_p, page_p, 1) ;// Macro to install basic funcs from page_p 

     int4                  header_size() const ;
     bool                  is_space_available() const ;

     int4                  data_start_offset() const ;
     void*                 data_start_addr() const ;

     int4                  bitmap_size() const ;
     int4                  bitmap_start_offset() const ;
     void*                 bitmap_start_addr() const ;
     void                  clear_bitmap() ; 

     static int4           max_data_records(int4 record_size) ; 

     int4                  nrecords() const ;
     int4                  nvacant() const ;
     smsize_t              record_size() const ;
     smsize_t              aligned_record_size() const ;
     shpid_t               next_extent() const ;
     rc_t                  link_extent( shpid_t ext ) ; 
     void*                 record_addr(const int4 idx) const ;
     bool                  is_record_valid(const int4 idx) const ;  

     bool                  is_fixed_p() const ; 
     rc_t                  link_up( shpid_t ppid, shpid_t npid ) ; 
      
     rc_t                  write_header( const fixed_p_hdr_t& hdr ) ; 
     rc_t                  insert_rec(const slotid_t idx, const vec_t& data) ;  
     rc_t                  insert_rec(const vec_t& data, slotid_t& idx) ; 
     rc_t                  update_rec(const slotid_t idx, 
                                      const smsize_t start, 
                                      const vec_t& data) ; 

     rc_t                  retrieve_rec(const slotid_t idx, 
                                        const smsize_t start, 
                                        uint4& len, void*& data) ; 

     rc_t                  retrieve_rec(const slotid_t idx, void*& data) ; 
     void                  retrieve_scan_rec(const slotid_t idx, void*& data) ; 

     int4                  next_rec(const slotid_t idx) ; 

     rc_t                  delete_rec(const slotid_t idx) ; 
     rc_t                  delete_rec(const slotid_t idx, void*& data) ; 
} ;

inline int4 fixed_p::nvacant() const
{
   return _pp->nvacant ;
}

inline int4 fixed_p::header_size() const
{
   return sizeof(fixed_p_hdr_t) ;
}

inline bool fixed_p::is_space_available() const
{
   return nvacant() > 0  ;
}

inline int4 fixed_p::nrecords() const
{
   return page_p::nslots() ;
}

inline smsize_t fixed_p::record_size() const
{
   return ((fixed_p_hdr_t *)_pp->data)->actual_record_size ;
}

inline shpid_t fixed_p::next_extent() const
{
   return ((fixed_p_hdr_t *)_pp->data)->next_ext ;
}

inline rc_t fixed_p::link_extent(shpid_t ext)
{
   ((fixed_p_hdr_t *)_pp->data)->next_ext = ext ; 
   return RCOK ; 
}

inline smsize_t fixed_p::aligned_record_size() const
{
   return align(record_size()) ;
}

inline int4 fixed_p::bitmap_size() const
{
   return data_sz - nrecords() * aligned_record_size() ; 
}

inline int4 fixed_p::bitmap_start_offset() const
{
   return page_p::hdr_size() + header_size() + data_sz - bitmap_size() ;
}

inline void* fixed_p::bitmap_start_addr() const
{
   return _pp->data + header_size() + data_sz - bitmap_size() ;
}

inline int4 fixed_p::data_start_offset() const
{
   // Sum of page header and header for fixed_p
   return page_p::hdr_size() + header_size() ;
}

inline void* fixed_p::data_start_addr() const
{
   return _pp->data +  header_size() ;
}

inline void* fixed_p::record_addr(const int4 idx) const
{
   return data_start_addr() + idx * aligned_record_size() ;
}

inline bool fixed_p::is_record_valid(const int4 idx) const
{
   return bm_is_set( (u_char *)bitmap_start_addr(), idx ) ;
}

inline bool fixed_p::is_fixed_p() const 
{
   // all pages in file must be t_fixed
   w_assert3(tag()&t_fixed_p);
   return tag()&t_fixed_p;
}

inline rc_t fixed_p::link_up( shpid_t ppid, shpid_t npid )
{
    return page_p::link_up(ppid, npid) ; 
}

inline void fixed_p::retrieve_scan_rec(const slotid_t idx, void*& data)  
{
    data = record_addr( idx ) ; 
}

inline int4 fixed_p::next_rec(const slotid_t idx)
{
    if ( idx > nrecords() - 2 ) return -1 ;
    return bm_first_set( (u_char*)bitmap_start_addr(), nrecords(), idx+1 ) ;
}

inline int4 fixed_p::max_data_records(int4 record_size)
{
   int4 tmp = mul8(align(record_size)) + 1 ;

   return mul8(fixed_p::data_sz - 1) / tmp ;
}

// Write the header on the page
inline rc_t fixed_p::write_header( const fixed_p_hdr_t& hdr )
{
   memcpy( _pp->data, (void *)&hdr, sizeof(hdr) ) ; 
   return RCOK ;  
}

#endif
